import sys
import os
import socket
import subprocess
import paramiko
from scp import SCPClient
import logging
import requests
import time

try:
    from .__init__ import __version__
except (ImportError,ValueError):
    from twigs.__init__ import __version__

GoDaddyCABundle = True
run_args = None

SYSTEM_TAGS = ['IMAGE_NAME', 'OS_VERSION', 'OS_RELEASE', 'SOURCE', 'OS_RELEASE_ID', 'CRITICALITY', 'http', 'https', 'OS_ARCH', 'IMAGE_DIGEST', 'DISCOVERY_TYPE']

def get_ssh_client(host):
    client = paramiko.SSHClient()
    client.set_missing_host_key_policy(paramiko.client.AutoAddPolicy)
    if host.get('userpwd') is not None and len(host['userpwd']) > 0 and (host.get('privatekey') is None or len(host['privatekey'])==0):
        client.connect(host['hostname'],username=host['userlogin'],password=host['userpwd'])
    elif host.get('privatekey') is not None and len(host['privatekey']) > 0:
        if host.get('userpwd') is not None and len(host['userpwd']) > 0:
            client.connect(host['hostname'],username=host['userlogin'],key_filename=host['privatekey'],passphrase=host['userpwd'])
        else:
            client.connect(host['hostname'],username=host['userlogin'],key_filename=host['privatekey'])
    else:
        client.connect(host['hostname'],username=host['userlogin'])
    return client

def run_cmd_on_host(args, host, cmdarr, logging_enabled=True):
    if host and host['remote']:
        pkgout = run_remote_ssh_command(args, host, cmdarr[0])
        if pkgout is None:
            return None
    else:
        try:
            dev_null_device = open(os.devnull, "w")
            pkgout = subprocess.check_output(cmdarr, stderr=dev_null_device, shell=True)
            pkgout = pkgout.decode(args.encoding)
            dev_null_device.close()
        except subprocess.CalledProcessError:
            if logging_enabled:
                logging.debug("Error running command")
            return None
    return pkgout

def run_remote_ssh_command_helper(client, command, args):
    logging.debug("Running command [%s] on remote host", command)
    output = ''
    stdin, stdout, stderr = client.exec_command(command)
    for line in stdout:
        output = output + line
    logging.debug("Command returned output:\n%s", output)
    return output, stdout.channel.recv_exit_status()

def run_remote_ssh_command(args, host, command):
    assetid = host['assetid'] if host.get('assetid') is not None else host['hostname']
    try:
        client = get_ssh_client(host)
        output, exit_code = run_remote_ssh_command_helper(client, command, args)
        client.close()
    except paramiko.ssh_exception.AuthenticationException as e:
        logging.info("Authentication failed for asset [%s], host [%s]", assetid, host['hostname'])
        logging.info("Exception: %s", e)
        output = None
    except paramiko.ssh_exception.SSHException as e:
        logging.info("SSHException while connecting to asset [%s], host [%s]", assetid, host['hostname'])
        logging.info("Exception: %s", e)
        output = None
    except socket.error as e:
        logging.info("Socket error while connection to asset [%s], host [%s]", assetid, host['hostname'])
        logging.info("Exception: %s", e)
        output = None
    except:
        logging.info("Unknown error running remote discovery for asset [%s], host [%s]: [%s]", assetid, host['hostname'], sys.exc_info()[0])
        output = None
    finally:
        return output

def run_script_on_host(args, host, script_path, script_args):
    if host and host['remote']:
        pkgout, exit_code = run_remote_ssh_script(args, host, script_path, script_args)
    else:
        if not os.path.isfile(script_path):
            logging.error("Error script file not found at [%s]", script_path)
            return None
        if not os.access(script_path, os.X_OK):
            logging.error("Error script file at [%s] is not marked as executable", script_path)
            return None
        try:
            dev_null_device = open(os.devnull, "w")
            #cmdarr = [ "chroot '" + root_folder + "' && " + script_path ]
            cmd = script_path
            for sa in script_args:
                cmd = cmd + ' "' + sa + '"'
            cmdarr = [ cmd ]
            pkgout = subprocess.check_output(cmdarr, stderr=dev_null_device, shell=True)
            pkgout = pkgout.decode(args.encoding)
            dev_null_device.close()
            exit_code = 0
        except subprocess.CalledProcessError as excpt:
            logging.error("Error running script [%s]", script_path)
            pkgout = None
            exit_code = excpt.returncode
    return pkgout, exit_code

def scp_put_file(host, from_path, to_path):
    _scp_file_helper(host, from_path, to_path, True)

def scp_get_file(host, from_path, to_path):
    _scp_file_helper(host, from_path, to_path, False)

def _scp_file_helper(host, from_path, to_path, put_flag=True):
    assetid = host['assetid'] if host.get('assetid') is not None else host['hostname']
    ret = False
    try:
        ssh_client = get_ssh_client(host)
        scp_client = SCPClient(ssh_client.get_transport())
        if put_flag == True: # put file to remote machine
            scp_client.put(from_path, recursive=True, remote_path=to_path)
        else:
            scp_client.get(from_path, to_path)
        scp_client.close()
        ssh_client.close()
        ret = True
    except paramiko.ssh_exception.AuthenticationException as e:
        logging.info("Authentication failed for asset [%s], host [%s]", assetid, host['hostname'])
        logging.info("Exception: %s", e)
    except paramiko.ssh_exception.SSHException as e:
        logging.info("SSHException while connecting to asset [%s], host [%s]", assetid, host['hostname'])
        logging.info("Exception: %s", e)
    except socket.error as e:
        logging.info("Socket error while connection to asset [%s], host [%s]", assetid, host['hostname'])
        logging.info("Exception: %s", e)
    except:
        logging.info("Unknown error running remote discovery for asset [%s], host [%s]: [%s]", assetid, host['hostname'], sys.exc_info()[0])
    finally:
        return ret

# NOTE script_path is path of script file on remote host
def run_remote_ssh_script(args, host, script_path, script_args):
    assetid = host['assetid'] if host.get('assetid') is not None else host['hostname']
    output = ''
    exit_code = -1
    try:
        client = get_ssh_client(host)
        cmd = "test -x " + script_path
        temp_output, exit_code = run_remote_ssh_command_helper(client, cmd, args)
        if exit_code != 0:
            logging.error("Error executable script file not found on remote host at [%s]", script_path)
            output = None, exit_code
        cmd = script_path
        for sa in script_args:
            cmd = cmd + ' "' + sa + '"'
        output, exit_code = run_remote_ssh_command_helper(client, cmd, args)
        client.close()
    except paramiko.ssh_exception.AuthenticationException as e:
        logging.info("Authentication failed for asset [%s], host [%s]", assetid, host['hostname'])
        logging.info("Exception: %s", e)
        output = None
    except paramiko.ssh_exception.SSHException as e:
        logging.info("SSHException while connecting to asset [%s], host [%s]", assetid, host['hostname'])
        logging.info("Exception: %s", e)
        output = None
    except socket.error as e:
        logging.info("Socket error while connection to asset [%s], host [%s]", assetid, host['hostname'])
        logging.info("Exception: %s", e)
        output = None
    except:
        logging.info("Unknown error running remote discovery for asset [%s], host [%s]: [%s]", assetid, host['hostname'], sys.exc_info()[0])
        output = None
    finally:
        return output, exit_code

def get_os_release(args, host=None):
    freebsd = False
    out = None
    cmdarr = ["/bin/cat /etc/os-release"]
    out = run_cmd_on_host(args, host, cmdarr, False)

    if out is None or out.strip() == '':

        # try redhat-release
        cmdarr = ["/bin/cat /etc/redhat-release"]
        out = run_cmd_on_host(args, host, cmdarr, False)
        if out is not None and out.strip() != '':
            return out.strip()
        else:
            # try FreeBSD
            cmdarr = ["/usr/bin/uname -v -p"]
            out = run_cmd_on_host(args, host, cmdarr, False)

            if out is not None and 'FreeBSD' not in out:
                # try OpenBSD
                cmdarr = ["/usr/bin/uname -srvm"]
                out = run_cmd_on_host(args, host, cmdarr, False)

    if out is None:
        logging.error("Failed to get os-release")
        return None

    if 'FreeBSD' in out or 'OpenBSD' in out:
        return out
    elif 'Darwin' in out:
        # Check for Mac OS
        cmdarr = ["sw_vers"]
        out = run_cmd_on_host(args, host, cmdarr, False)
        if out is not None and out.strip() != '':
            mac_os_version = ""
            for line in out.splitlines():
                value = line.split(':')[1].strip()
                mac_os_version = value if len(mac_os_version) == 0 else mac_os_version + " " + value
            return mac_os_version
    else:
        output_lines = out.splitlines()
        for l in output_lines:
            if 'PRETTY_NAME' in l:
                return l.split('=')[1].replace('"','')
    return None

def get_asset_type(os):
    os = os.lower()
    if "centos" in os:
        return "CentOS"
    elif "red hat" in os:
        return "Red Hat"
    elif "ubuntu" in os:
        return "Ubuntu"
    elif "debian" in os:
        return "Debian"
    elif "amazon linux" in os:
        return "Amazon Linux AMI"
    elif "oracle linux" in os:
        return "Oracle Linux"
    elif "freebsd" in os:
        return "FreeBSD"
    elif "openbsd" in os:
        return "OpenBSD"
    elif "suse" in os:
        return "Suse"
    elif "mac os" in os or "macos" in os:
        return "Mac OS"
    elif "windows" in os:
        return "Windows"
    elif "alpine" in os:
        return "Alpine Linux"
    elif "container-optimized os" in os:
        return "Google Container-Optimized OS"
    elif "rocky linux" in os:
        return "Rocky Linux"
    else:
        logging.error("Not a supported OS type [%s]" % os)
        return None

# reference: https://gist.github.com/bencord0/7690953
def get_unique_asset_id(args, host, asset_type):

    # Try to get MAC Address
    output = get_mac_address(args, host, asset_type)
    if output is not None and len(output.strip()) > 0:
        return output

    output = run_cmd_on_host(args, host, ["cat /var/lib/dbus/machine-id"], False)
    if output is not None and len(output.strip()) > 0:
        return output.strip()

    output = run_cmd_on_host(args, host, ["cat /sys/class/dmi/id/product_uuid"], False)
    if output is not None and len(output.strip()) > 0:
        return output.strip()

    # For FreeBSD
    output = run_cmd_on_host(args, host, ["sysctl kern.hostuuid"], False)
    if output is not None and len(output.strip()) > 0:
        return output.strip().split(' ')[1]

    # For Mac OS
    output = run_cmd_on_host(args, host, ["sysctl kern.uuid"], False)
    if output is not None and len(output.strip()) > 0:
        return output.strip().split(' ')[1]


def get_mac_address(args, host, asset_type):

    # get default interface
    cmd = "ip route show default | awk '/default/ {print $5}'"
    defif = run_cmd_on_host(args, host, [cmd], True)
    if defif == None or defif.strip() == '':
        return None
    defif = defif.strip()

    # now list mac address from /sys/class/net
    cmd = "cat /sys/class/net/"+defif+"/address"
    macid = run_cmd_on_host(args, host, [cmd], True)
    if macid == None or macid.strip() == '':
        return None
    macid = macid.strip().replace(':','')
    return macid 

def get_ip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        # doesn't even have to be reachable
        s.connect(('10.255.255.255', 1))
        IP = s.getsockname()[0]
    except:
        IP = '127.0.0.1'
    finally:
        s.close()
    return IP

def find_files(localpath, filename):
    ret_files = []
    for root, subdirs, files in os.walk(localpath):
        for fname in files:
            file_path = os.path.join(root, fname)
            if len(filename) == 0:
                ret_files.append(file_path)
            elif file_path.endswith(filename):
                ret_files.append(file_path)
    return ret_files

def ascii_string(in_str):
    ascii_str = ''.join([c if ord(c) < 128 else ' ' for c in in_str.strip()])
    return ascii_str

def get_indent(line):
    return len(line) - len(line.lstrip(' '))

def get_rating(cvss_score):
    rating = '1'
    if cvss_score is None or cvss_score == '':
        return rating
    s = float(cvss_score)
    if s <= 2:
        rating = '1'
    elif s <= 4:
        rating = '2'
    elif s <= 6:
        rating = '3'
    elif s <= 8:
        rating = '4'
    elif s <= 10:
        rating = '5'
    return rating

def set_requests_verify(verify):
    global GoDaddyCABundle
    GoDaddyCABundle = verify

def get_requests_verify():
    global GoDaddyCABundle
    return GoDaddyCABundle

def requests_get(url):
    rc = 0
    st = 1
    while True:
        try:
            resp = requests.get(url, verify=get_requests_verify())
            resp_status_code = resp.status_code
        except requests.exceptions.RequestException as e:
            logging.warning("Retry count [%s] got exception: [%s]", rc, str(e))
            if rc >= 10:
                logging.warning("Max retries exceeded....giving up...")
                return None
            else:
                logging.warning("Sleeping for [%s] seconds...", st)
                time.sleep(st)
                rc = rc + 1
                st = st * 2
                continue
        return resp

def requests_post(url, json):
    rc = 0
    st = 1
    while True:
        try:
            resp =  requests.post(url, json=json, verify=get_requests_verify())
            resp_status_code = resp.status_code
        except requests.exceptions.RequestException as e:
            logging.warning("Retry count [%s] got exception: [%s]", rc, str(e))
            if rc >= 10:
                logging.warning("Max retries exceeded....giving up...")
                return None
            else:
                logging.warning("Sleeping for [%s] seconds...", st)
                time.sleep(st)
                rc = rc + 1
                st = st * 2
                continue
        return resp

def requests_post_files(url, files):
    rc = 0
    st = 1
    while True:
        try:
            resp =  requests.post(url, files=files, verify=get_requests_verify())
            resp_status_code = resp.status_code
        except requests.exceptions.RequestException as e:
            logging.warn("Retry count [%s] got exception: [%s]", rc, str(e))
            if rc >= 10:
                logging.warn("Max retries exceeded....giving up...")
                return None
            else:
                logging.warn("Sleeping for [%s] seconds...", st)
                time.sleep(st)
                rc = rc + 1
                st = st * 2
                continue
        return resp

def requests_put(url, json):
    rc = 0
    st = 1
    while True:
        try:
            resp = requests.put(url, json=json, verify=get_requests_verify())
            resp_status_code = resp.status_code
        except requests.exceptions.RequestException as e:
            logging.warning("Retry count [%s] got exception: [%s]", rc, str(e))
            if rc >= 10:
                logging.warning("Max retries exceeded....giving up...")
                return None
            else:
                logging.warning("Sleeping for [%s] seconds...", st)
                time.sleep(st)
                rc = rc + 1
                st = st * 2
                continue
        return resp

def get_asset(asset_id, args):
    asset_url = "https://" + args.instance + "/api/v2/assets/"
    auth_data = "?handle=" + args.handle + "&token=" + args.token + "&format=json"

    resp = requests_get(asset_url + asset_id + "/" + auth_data)
    if resp is not None and resp.status_code == 200:
        return resp.json()
    else:
        return None

def tw_open(in_file, in_encoding, in_mode='rt'):
    try:
        if sys.version_info[0] < 3:
            f = open(in_file, mode=in_mode)
        else:
            f = open(in_file, mode=in_mode, encoding=in_encoding)
        return f
    except Exception as e:
        logging.warning("Unable to open file ["+in_file+"]")
        return None

def set_run_args(args):
    global run_args
    run_args = args

def get_run_args():
    global run_args
    return run_args

def create_new_tool_run_record():
    args = get_run_args()
    if args.token is None or len(args.token) == 0:
        return None
    run_id  = args.mode if args.run_id is None else args.run_id
    machine_id = get_ip()
    tool_name = 'twigs'
    tool_version = __version__
    email = args.handle
    url = "https://" + args.instance + "/api/v1/toolrun/new/"
    auth_data = "?handle=" + args.handle + "&token=" + args.token + "&format=json"
    post_payload = { "run_id": run_id, "machine_id": machine_id, "email": email, "name": tool_name, "version": tool_version }
    response = requests_post(url + auth_data, post_payload)
    if response.status_code == 200:
        args.run_record_id = response.json()['id']
    return response

def update_tool_run_record(status='RUNNING', observations=None):
    args = get_run_args()
    if args.token is None or len(args.token) == 0:
        return None
    run_record_id = args.run_record_id
    url = "https://" + args.instance + "/api/v1/toolrun/" + str(run_record_id) + "/"
    auth_data = "?handle=" + args.handle + "&token=" + args.token + "&format=json"
    if status != 'RUNNING':
        logging.shutdown()
        lfn_suffix = args.run_id if args.run_id is not None else args.mode
        logfilename = "twigs_%s.log" % lfn_suffix
        observations = { } if observations is None else observations
        observations['log_snippet'] = tail(logfilename, args.encoding, 100)
    put_payload = { "status": status, "observations": observations }
    response = requests_put(url + auth_data, put_payload)
    return response

def tw_exit(exit_code, observations = None):
    status = 'SUCCESS' if exit_code == 0 else 'FAIL'
    update_tool_run_record(status, observations)
    sys.exit(exit_code)

def tail(fn, encoding, lines=20):
    with open(fn, 'rb') as f:
        total_lines_wanted = lines

        BLOCK_SIZE = 1024
        f.seek(0, 2)
        block_end_byte = f.tell()
        lines_to_go = total_lines_wanted
        block_number = -1
        blocks = [] # blocks of size BLOCK_SIZE, in reverse order starting
                    # from the end of the file
        while lines_to_go > 0 and block_end_byte > 0:
            if (block_end_byte - BLOCK_SIZE > 0):
                # read the last block we haven't yet read
                f.seek(block_number*BLOCK_SIZE, 2)
                blocks.append(f.read(BLOCK_SIZE).decode(encoding))
            else:
                # file too small, start from begining
                f.seek(0,0)
                # only read what was not read
                blocks.append(f.read(block_end_byte).decode(encoding))
            lines_found = blocks[-1].count('\n')
            lines_to_go -= lines_found
            block_end_byte -= BLOCK_SIZE
            block_number -= 1
        all_read_text = ''.join(reversed(blocks))
        return '\n'.join(all_read_text.splitlines()[-total_lines_wanted:])

def get_latest_version():
    response = requests.get('https://pypi.org/pypi/twigs/json')
    latest_version = response.json()['info']['version']
    return latest_version




import sys
import json
import os
import requests
import re
import logging
import random
import time

from . import utils

def get_organizations(args, headers):
   url = args.base_url + f'/organizations'
   resp = requests.get(url, headers=headers)
   resp.raise_for_status()
   return resp.json()

def get_networks(args, headers, org_id):
   url = args.base_url + f'/organizations/{org_id}/networks'
   resp = requests.get(url, headers=headers, verify=False)
   resp.raise_for_status()
   return resp.json()

def get_devices(args, headers, org_id):
   url = args.base_url + f'/organizations/{org_id}/devices'
   resp = requests.get(url, headers=headers, verify=False)
   resp.raise_for_status()
   return resp.json()

def get_firmware_upgrades(args, headers, network_id):
   url = args.base_url + f'/networks/{network_id}/firmwareUpgrades'
   resp = requests.get(url, headers=headers, verify=False)
   if resp.status_code == 404:
       return None  # Some networks may not support this endpoint
   resp.raise_for_status()
   return resp.json()

def get_sm_devices(args, headers, network_id):
   url = args.base_url + f'/networks/{network_id}/sm/devices'
   logging.info("SM devices url: "+url)
   resp = requests.get(url, headers=headers, verify=False)
   if resp.status_code == 404:
       logging.info("SM devices response status: "+str(resp.status_code))
       logging.info("SM devices response: "+resp.text)
       return None  # Some networks may not support this endpoint
   resp.raise_for_status()
   return resp.json()

def get_sm_device_softwares(args, headers, network_id, device_id):
   url = args.base_url + f'/networks/{network_id}/sm/devices/{device_id}/softwares'
   resp = requests.get(url, headers=headers, verify=False)
   if resp.status_code == 404:
       return None  # Some networks may not support this endpoint
   resp.raise_for_status()
   return resp.json()

def get_all_devices(args, headers):
    assets = []
    logging.info("Retrieving network inventory from Cisco Meraki") 
    organizations = get_organizations(args, headers)
    for org in organizations:
        logging.info(f"Organization: {org['name']} ({org['id']})")
        networks = get_networks(args, headers, org['id'])
        devices = get_devices(args, headers, org['id'])
        devices_by_network = {}
        for device in devices:
            devices_by_network.setdefault(device['networkId'], []).append(device)
        for net in networks:
            logging.info(f"Network: {net['name']} ({net['id']})")
            fw_info = get_firmware_upgrades(args, headers, net['id'])
            fw_versions = {}
            if fw_info and 'products' in fw_info:
               for product in fw_info['products']:
                   fw_versions[product] = fw_info['products'][product]['currentVersion']['shortName']
            for device in devices_by_network.get(net['id'], []):
                product_type = device.get('productType', 'unknown')
                fw_version = fw_versions.get(product_type, 'unknown')
                logging.info(f"Device: {device.get('name', device['serial'])} | Model: {device['model']} | Product: {product_type} | Firmware: {fw_version}")
                asset = {}
                asset['owner'] = args.handle
                asset['id'] = device['serial']
                asset['name'] = device['name']
                asset['type'] = 'Cisco'
                asset_tags = ['Cisco Meraki']
                asset['tags'] = asset_tags
                model = device['model']
                if '-' in model:
                    model = model.split('-')[0]
                prodstr = 'Cisco Meraki '+model
                if fw_version != None:
                    prodstr = prodstr + ' firmware ' + fw_version.split()[1]
                products = [prodstr]
                asset['products'] = products
                assets.append(asset)
            sm_devices = get_sm_devices(args, headers, net['id'])
            if not sm_devices:
                logging.info(f"No SM devices for Network: {net['name']} ({net['id']})")
                continue
            for sm_device in sm_devices:
                jd = json.dumps(sm_device, indent=4)
                logging.info('SM Device')
                logging.info(jd)
                asset = {}
                asset['owner'] = args.handle
                asset['id'] = sm_device['id']
                asset['name'] = sm_device['name']
                if 'osName' in sm_device and 'Android' in sm_device['osName']:
                    asset['type'] = 'Android'
                elif 'osName' in sm_device and 'iOS' in sm_device['osName']:
                    asset['type'] = 'iPhone'
                asset_tags = list(set(['Cisco Meraki MDM', asset['type']]))
                if 'tags' in sm_device:
                    for t in sm_device['tags']:
                        asset_tags.append(t)
                asset['tags'] = asset_tags
                if 'systemModel' in sm_device and sm_device['systemModel'].startswith('SM-'):
                    # samsung android - adjust the version number to match dot decimal
                    products = ['Samsung '+sm_device['osName']+'.0']
                else:
                    products = [sm_device['osName']]
                softwares = get_sm_device_softwares(args, headers, net['id'], sm_device['id'])
                for software in softwares:
                    js = json.dumps(software, indent=4)
                    logging.info(js)
                    prodstr = software['identifier'] + ' ' + software['shortVersion'] 
                    products.append(prodstr)
                    vendor = ''
                    if 'vendor' in software and software['vendor'] != 'Unknown':
                        vendor = software['vendor']
                        vendor = vendor.replace('Corporation','')
                        vendor = vendor.replace('LLC','')
                        vendor = vendor.strip()
                    prodstr = vendor + ' ' + software['name'] + ' ' + software['shortVersion'] 
                    products.append(prodstr)
                asset['products'] = products
                assets.append(asset)
    return assets

# Main entry point
def get_inventory(args):
    headers = {
        'X-Cisco-Meraki-API-Key': args.api_key,
        'Content-Type': 'application/json',
        'Accept': 'application/json'
    }
    return get_all_devices(args, headers)




import sys
import os
import logging
import json
import tempfile
import shutil
from . import utils as lib_utils
from . import repo
from . import code_secrets as lib_code_secrets
from . import sast
from . import iac

def get_inventory(args):
    ret_assets = []
    cmd = "az functionapp list"
    out = lib_utils.run_cmd_on_host(args, None, cmd)
    fadict = json.loads(out)
    for fa in fadict:
        logging.info("Discovering Azure FunctionApp - "+fa['name'])
        faid = fa['id']
        cmd = "az functionapp deployment list-publishing-profiles --ids "+faid
        out = lib_utils.run_cmd_on_host(args, None, cmd)
        fadeploy = json.loads(out)
        for fad in fadeploy:
            basepath = None
            if fad['publishMethod'] == 'FTP':
                logging.info("Discovering Azure Function")
                assets = []
                temp_dir = tempfile.mkdtemp()
                user = fad['userName'].split('\\')[1].replace('$','')
                passwd = fad['userPWD']
                ftpurl = fad['publishUrl']
                cmd = 'wget --timeout=5 -m --user '+user+' --password '+passwd+' '+ftpurl+' -P '+temp_dir
                out = lib_utils.run_cmd_on_host(args, None, cmd)
                basepath = temp_dir + '/' + ftpurl.replace('ftp://','') 
                flist = [os.path.join(basepath, o) for o in os.listdir(basepath)
                    if os.path.isdir(os.path.join(basepath,o))]
                for fdir in flist:
                    fname = fdir.split('/')[-1]
                    args.assetid = fa['id'] + '-' + fname
                    args.assetname = fa['name'] + '-' + fname
                    args.repo = fdir
                    path = basepath + '/' + fname
                    logging.info("Discovering code dependencies for vulnerability scan.")
                    assets = repo.discover_inventory(args, path)
                    if args.secrets_scan:
                        logging.info("Discovering secrets/sensitive information. This may take some time.")
                        secret_records = lib_code_secrets.scan_for_secrets(args, path, basepath)
                        assets[0]['secrets'] = secret_records

                    code_issues = []
                    if args.sast:
                        logging.info("Performing static analysis. This may take some time.")
                        sast_records = sast.run_sast(args, path, basepath)
                        code_issues.extend(sast_records)

                    if args.iac_checks:
                        logging.info("Identifying infrastructure as code (IaC) issues. This may take some time.")
                        iac_records = iac.run_iac_checks(args, path, basepath)
                        code_issues.extend(iac_records)

                    if len(code_issues) > 0:
                        assets[0]['sast'] = code_issues
                    assets[0]['type'] = 'Azure Function'
                    assets[0]['tags'].extend(['Azure', 'Azure Function', 'Serverless'])
                    ret_assets.extend(assets)
                if len(assets) == 0:
                    dataurl = ftpurl.replace('site/wwwroot','data/SitePackages')
                    cmd = 'wget --timeout=5 -m --user '+user+' --password '+passwd+' '+dataurl+' -P '+temp_dir
                    logging.info("Trying SitePackages - "+cmd)
                    logging.debug(cmd)
                    out = lib_utils.run_cmd_on_host(args, None, cmd)
                    if out == None:
                        logging.debug("No SitePackages found")
                        continue
                    basepath = temp_dir + '/' + dataurl.replace('ftp://','') 
                    args.assetid = fa['id']
                    args.assetname = fa['name']
                    args.repo = basepath.replace(temp_dir,'')
                    path = basepath
                    cmd = 'cat '+basepath+'/'+'packagename.txt'
                    logging.debug(cmd)
                    packagename = lib_utils.run_cmd_on_host(args, None, cmd)
                    if packagename == None:
                        logging.debug("Error getting function package name")
                        continue
                    cmd = 'cd '+basepath+';unzip '+packagename
                    logging.debug(cmd)
                    out = lib_utils.run_cmd_on_host(args, None, cmd)
                    if out == None:
                        logging.debug("Error getting function package contents")
                        continue 
                    logging.info("Discovering code dependencies for vulnerability scan.")
                    assets = repo.discover_inventory(args, path, basepath)
                    assets[0]['type'] = 'Azure Function'
                    assets[0]['tags'].extend(['Azure', 'Azure Function', 'Serverless'])
                    ret_assets.extend(assets)
                shutil.rmtree(temp_dir)
    return ret_assets 



import os
import logging

from . import utils
from . import plugin_registry

supported_field_names = ["TYPE", "VULN", "PERCENTAGE", "AFFECTED_PRODUCT", "VULNERABLE_PRODUCT", "ANALYSIS", "RECOMMENDATION"]
field_name_mapping = {
    "IMPACT": {
        "TYPE": "type",
        "VULN": "id_str",
        "PERCENTAGE": "percentage",
        "AFFECTED_PRODUCT": "keyword",
        "VULNERABLE_PRODUCT": "product",
        "ANALYSIS": "analysis",
        "RECOMMENDATION": "recommendation"
    }
}
required_fields_by_type = {
    "IMPACT": ["VULN", "PERCENTAGE", "AFFECTED_PRODUCT", "VULNERABLE_PRODUCT", "ANALYSIS"]
}

def validate_record(rtype, finding_dict):
    global required_fields_by_type
    required_fields = required_fields_by_type.get(rtype)
    if required_fields is None:
        logging.error("Plugin returned malformed response - invalid record type [%s]", rtype)
        return False
    for rf in required_fields:
        if finding_dict.get(rf) is None:
            logging.error("Plugin returned malformed response - missing required field [%s]", rf)
            return False
    return True

def transform_record(rtype, finding_dict):
    global field_name_mapping
    ret_dict = { }
    for key in finding_dict.keys():
        ret_dict[field_name_mapping[rtype][key]] = finding_dict[key]
    return ret_dict

def process_record(rtype, finding_dict, asset_dict):
    if rtype != '':
        if validate_record(rtype, finding_dict):
            finding_dict = transform_record(rtype, finding_dict)
            if rtype == "IMPACT":
                if asset_dict.get('impacts') is None:
                    asset_dict['impacts'] = []
                asset_dict['impacts'].append(finding_dict)
            else:
                logging.error("Plugin returned unsupported action!")

def process_plugin_output(plugin_output, asset_dict):
    global supported_field_names

    if plugin_output is None:
        return

    lines = plugin_output.splitlines()
    total_lines = len(lines)
    if total_lines == 0:
        return
    finding_dict = { }
    rtype = ''
    for cl in lines:
        if len(cl) == 0:
            process_record(rtype, finding_dict, asset_dict)
            rtype = ''
            finding_dict = { }
            continue
        if ':' not in cl:
            logging.error("Plugin returned malformed response - missing separator")
            break
        tokens = cl.split(':')
        if len(tokens) == 1:
            # there is no value specified, so skip the field
            continue
        if len(tokens) > 2:
            logging.error("Plugin returned malformed response - incorrect number of fields")
            break
        prefix = tokens[0]
        value = tokens[1]
        if prefix not in supported_field_names:
            logging.error("Plugin returned malformed response - unknown field type")
            break
        if prefix == "TYPE":
            # encountered new record, so first validate earlier record
            process_record(rtype, finding_dict, asset_dict)
            rtype = value
            finding_dict = { }
        else:
            finding_dict[prefix] = value
    if rtype != '' and len(finding_dict) > 0:
        process_record(rtype, finding_dict, asset_dict)

def execute_plugin(args, asset_dict, host, plugin_dict, root_folder):
    plugin_dir = plugin_dict['plugin_dir']
    plugin_abs_path = plugin_dir + os.sep + plugin_dict.get('file')
    logging.info("Running plugin [%s]", plugin_dict['name'])
    script_output, exit_code = utils.run_script_on_host(args, host, plugin_abs_path, [root_folder])

    if exit_code == 0:
        logging.info("Processing output from plugin [%s]", plugin_dict['name'])
        process_plugin_output(script_output, asset_dict)
        utils.update_tool_run_record()
    elif exit_code == 1:
        logging.error("Plugin [%s] failed to run successfully", plugin_dict['name'])

def run_plugin(args, asset_dict, host, these_checks, plugin_dir, root_folder):
    pr = plugin_registry.get_plugin_registry()
    for this_check in these_checks:
        plugin = pr.get(this_check.lower())
        if plugin is None:
            logging.error("No plugin found for [%s]", this_check)
            continue
        else:
            if plugin.get('enabled'):
                plugin['plugin_dir'] = plugin_dir
                execute_plugin(args, asset_dict, host, plugin, root_folder)
            else:
                logging.error("Specified plugin [%s] is not enabled", this_plugin)
                continue

def run_plugins(args, asset_dict, host, plugin_dir, root_folder):
    logging.info("Running all enabled plugins")
    pr = plugin_registry.get_plugin_registry()
    plugins = pr.keys()
    for p in plugins:
        if pr[p]['enabled'] == False:
            continue
        pr[p]['plugin_dir'] = plugin_dir
        execute_plugin(args, asset_dict, host, pr[p], root_folder)
    logging.info("Done running all enabled plugins")

def process_plugins(asset_dict, host, args, root_folder):
    if args.mode not in ['host', 'docker', 'gcr', 'acr', 'ecr', 'ocr', 'k8s']:
        return

    assetid = host['assetid'] if host.get('assetid') is not None else host['hostname']
    try:
        plugin_dir = plugin_registry.get_plugin_dir()
        if host['remote'] == True:
            # copy plugins folder on remote host
            client = utils.get_ssh_client(host)
            cmd = 'mktemp -d'
            remote_plugin_dir, exit_code = utils.run_remote_ssh_command_helper(client, cmd, args)
            remote_plugin_dir = remote_plugin_dir.strip()
            utils.scp_put_file(client, plugin_dir, remote_plugin_dir)
            cmd = 'chmod +x ' + remote_plugin_dir + '/*'
            utils.run_remote_ssh_command_helper(client, cmd, args)
            plugin_dir = remote_plugin_dir + os.sep + 'plugins'
            client.close()

        if args.check_vuln is not None:
            run_plugin(args, asset_dict, host, args.check_vuln, plugin_dir, root_folder)
        elif args.check_all_vulns:
            run_plugins(args, asset_dict, host, plugin_dir, root_folder)

        if host['remote'] == True:
            # remove plugin dir from remote host
            client = utils.get_ssh_client(host)
            cmd = 'rm -rf ' + remote_plugin_dir
            utils.run_remote_ssh_command_helper(client, cmd, args)
            client.close()
    except paramiko.ssh_exception.AuthenticationException as e:
        logging.info("Authentication failed for asset [%s], host [%s]", assetid, host['hostname'])
        logging.info("Exception: %s", e)
    except paramiko.ssh_exception.SSHException as e:
        logging.info("SSHException while connecting to asset [%s], host [%s]", assetid, host['hostname'])
        logging.info("Exception: %s", e)
    except socket.error as e:
        logging.info("Socket error while connection to asset [%s], host [%s]", assetid, host['hostname'])
        logging.info("Exception: %s", e)
    except:
        logging.info("Unknown error running remote discovery for asset [%s], host [%s]: [%s]", assetid, host['hostname'], sys.exc_info()[0])
    finally:
        return



import os
import logging
import json

from . import utils

SUPPORTED_SBOM_FORMATS_FOR_STANDARD = {
        "cyclonedx": ["json"],
        "spdx": ["tagvalue", "json"],
        "threatworx": ["json", "csv"] # keep this one last
}
SUPPORTED_SBOM_STANDARDS = list(SUPPORTED_SBOM_FORMATS_FOR_STANDARD.keys())

def upload_sbom(args):
    sbom_standard = args.standard
    sbom_format = args.format
    sbom_abs_path = os.path.abspath(args.input)

    if sbom_standard not in SUPPORTED_SBOM_STANDARDS:
        logging.error("Unsupported SBOM standard [%s]", sbom_standard)
        return False

    supported_sbom_formats = SUPPORTED_SBOM_FORMATS_FOR_STANDARD[sbom_standard]
    if sbom_format not in supported_sbom_formats:
        logging.error("Unsupported format [%s] for SBOM standard [%s]", sbom_format, sbom_standard)
        return False

    if os.path.isfile(sbom_abs_path) == False:
        logging.error("Unable to access SBOM file [%s]", sbom_abs_path)
        return False

    json_data = { }
    json_data['sbom_standard'] = sbom_standard
    json_data['sbom_format'] = sbom_format
    if args.assetid:
        json_data['asset_id'] = args.assetid
    if args.assetname:
        json_data['asset_name'] = args.assetname
    if args.org:
        json_data['org'] = args.org
    if args.tag:
        json_data['tags'] = args.tag
    if args.comment:
        json_data['comment'] = args.comment
    if args.apply_policy is not None:
        # If policy is specified, then no need to run separate scan
        json_data['run_scan'] = False
    else:
        json_data['run_scan'] = True

    sbom_upload_url = "https://" + args.instance + "/api/v2/assets/sbom/"
    auth_data = "?handle=" + args.handle + "&token=" + args.token + "&format=json"

    files = [
        ('sbom_artifact', ('sbom_artifact', open(sbom_abs_path, 'rb') )),
        ('data', ('data', json.dumps(json_data), 'application/json')),
    ]
    logging.info("Uploading SBOM artifact [%s]", sbom_abs_path)
    resp = utils.requests_post_files(sbom_upload_url + auth_data, files)
    if resp is not None and resp.status_code == 200:
        logging.info("Successfully uploaded SBOM artifact")
        resp_json = resp.json()
        return True, resp_json.get('asset_ids')
    else:
        logging.error("Uploading SBOM artifact failed")
        logging.error(resp.text)
        return False, None




import sys
import re
import os
import shutil
import stat
import subprocess
import logging
import json
import tempfile
import traceback

def get_rating(sev):
    if sev == 'INFO' or sev == 'OK':
        return '1'
    if sev == 'LOW':
        return '2'
    if sev == 'MEDIUM':
        return '3'
    if sev == 'HIGH':
        return '4'
    if sev == 'CRITICAL':
        return '5'
    return '1'

def run_ssl_audit(url, assetid):
    findings = []
    SSL_AUDIT_PATH = os.path.dirname(os.path.realpath(__file__)) + '/ssl_audit/testssl.sh'
    temp_name = next(tempfile._get_candidate_names())
    defult_tmp_dir = tempfile._get_default_tempdir()
    audit_out = defult_tmp_dir + '/' + temp_name + '.json'
    try:
        cmd = SSL_AUDIT_PATH + ' -oJ ' + audit_out + ' ' +url
        cmdarr = [cmd]
        dev_null_device = open(os.devnull, "w")
        subprocess.check_output(cmdarr, stderr=dev_null_device, shell=True)
        dev_null_device.close()
    except subprocess.CalledProcessError as e:
        logging.debug("ssl audit error")
        logging.debug(str(e))
        return findings

    try:
        jf = open(audit_out, 'r')
        out = jf.read()
        jf.close()
        os.remove(audit_out)
        odict = json.loads(out)
    except Exception as e:
        logging.debug("error processing ssl audit output")
        logging.debug(str(e))
        return findings

    if 'scanResult' in odict:
        if 'pretest' in odict['scanResult'][0]:
            for p in odict['scanResult'][0]['pretest']:
                issue = {}
                issue['twc_id'] = 'ssl-audit-'+p['id']
                issue['twc_title'] = p['id']
                issue['details'] = p['finding']
                issue['rating'] = get_rating(p['severity']) 
                issue['object_id'] = url 
                issue['asset_id'] = assetid
                issue['object_meta'] = '' 
                issue['type'] = 'SSL' 
                findings.append(issue)
        if 'protocols' in odict['scanResult'][0]:
            for p in odict['scanResult'][0]['protocols']:
                issue = {}
                issue['twc_id'] = 'ssl-audit-'+p['id']
                issue['twc_title'] = 'protocol: '+p['id']
                issue['details'] = p['finding']
                issue['rating'] = get_rating(p['severity']) 
                issue['object_id'] = url 
                issue['asset_id'] = assetid
                issue['object_meta'] = '' 
                issue['type'] = 'SSL' 
                findings.append(issue)
        if 'grease' in odict['scanResult'][0]:
            for p in odict['scanResult'][0]['grease']:
                issue = {}
                issue['twc_id'] = 'ssl-audit-'+p['id']
                issue['twc_title'] = p['id']
                issue['details'] = p['finding']
                issue['rating'] = get_rating(p['severity']) 
                issue['object_id'] = url 
                issue['asset_id'] = assetid
                issue['object_meta'] = '' 
                issue['type'] = 'SSL' 
                findings.append(issue)
        if 'grease' in odict['scanResult'][0]:
            for p in odict['scanResult'][0]['grease']:
                issue = {}
                issue['twc_id'] = 'ssl-audit-'+p['id']
                issue['twc_title'] = p['id']
                issue['details'] = p['finding']
                issue['rating'] = get_rating(p['severity']) 
                issue['object_id'] = url 
                issue['asset_id'] = assetid
                issue['object_meta'] = '' 
                issue['type'] = 'SSL' 
                findings.append(issue)
        if 'ciphers' in odict['scanResult'][0]:
            for p in odict['scanResult'][0]['ciphers']:
                issue = {}
                issue['twc_id'] = 'ssl-audit-'+p['id']
                issue['twc_title'] = 'cipher: '+p['id']
                finding = p['finding']
                if 'cwe' in p:
                    finding = finding + ' ' + p['cwe']
                issue['details'] = finding
                issue['rating'] = get_rating(p['severity']) 
                issue['object_id'] = url 
                issue['asset_id'] = assetid
                issue['object_meta'] = ''
                issue['type'] = 'SSL' 
                findings.append(issue)
        if 'fs' in odict['scanResult'][0]:
            for p in odict['scanResult'][0]['fs']:
                issue = {}
                issue['twc_id'] = 'ssl-audit-'+p['id']
                issue['twc_title'] = 'forward secrecy: '+p['id']
                issue['details'] = p['finding']
                issue['rating'] = get_rating(p['severity']) 
                issue['object_id'] = url 
                issue['asset_id'] = assetid
                issue['object_meta'] = ''
                issue['type'] = 'SSL' 
                findings.append(issue)
        if 'serverPreferences' in odict['scanResult'][0]:
            for p in odict['scanResult'][0]['serverPreferences']:
                issue = {}
                issue['twc_id'] = 'ssl-audit-'+p['id']
                issue['twc_title'] = 'server preference: '+p['id']
                issue['details'] = p['finding']
                issue['rating'] = get_rating(p['severity']) 
                issue['object_id'] = url 
                issue['asset_id'] = assetid
                issue['object_meta'] = ''
                issue['type'] = 'SSL' 
                findings.append(issue)
        if 'serverDefaults' in odict['scanResult'][0]:
            for p in odict['scanResult'][0]['serverDefaults']:
                issue = {}
                issue['twc_id'] = 'ssl-audit-'+p['id']
                issue['twc_title'] = 'server default: '+p['id']
                issue['details'] = p['finding']
                issue['rating'] = get_rating(p['severity']) 
                issue['object_id'] = url 
                issue['asset_id'] = assetid
                issue['object_meta'] = ''
                issue['type'] = 'SSL' 
                findings.append(issue)
        if 'headerResponse' in odict['scanResult'][0]:
            for p in odict['scanResult'][0]['headerResponse']:
                issue = {}
                issue['twc_id'] = 'ssl-audit-'+p['id']
                issue['twc_title'] = 'header response: '+p['id']
                issue['details'] = p['finding']
                issue['rating'] = get_rating(p['severity']) 
                issue['object_id'] = url 
                issue['asset_id'] = assetid
                issue['object_meta'] = ''
                issue['type'] = 'SSL' 
                findings.append(issue)
        if 'vulnerabilities' in odict['scanResult'][0]:
            for p in odict['scanResult'][0]['vulnerabilities']:
                issue = {}
                issue['twc_id'] = 'ssl-audit-'+p['id']
                issue['twc_title'] = 'vulnerability: '+p['id']
                finding = p['finding']
                if 'cve' in p:
                    finding = finding + ' ' + p['cve']
                if 'cwe' in p:
                    finding = finding + ' ' + p['cwe']
                issue['details'] = finding
                issue['rating'] = get_rating(p['severity']) 
                issue['object_id'] = url 
                issue['asset_id'] = assetid
                issue['object_meta'] = ''
                issue['type'] = 'SSL' 
                findings.append(issue)
        if 'cipherTests' in odict['scanResult'][0]:
            for p in odict['scanResult'][0]['cipherTests']:
                issue = {}
                issue['twc_id'] = 'ssl-audit-'+p['id']
                issue['twc_title'] = 'cipher test: '+p['id']
                issue['details'] = p['finding']
                issue['rating'] = get_rating(p['severity']) 
                issue['object_id'] = url 
                issue['asset_id'] = assetid
                issue['object_meta'] = ''
                issue['type'] = 'SSL' 
                findings.append(issue)
        if 'browserSimulations' in odict['scanResult'][0]:
            for p in odict['scanResult'][0]['browserSimulations']:
                issue = {}
                issue['twc_id'] = 'ssl-audit-'+p['id']
                issue['twc_title'] = 'browser simulation: '+p['id']
                issue['details'] = p['finding']
                issue['rating'] = get_rating(p['severity']) 
                issue['object_id'] = url 
                issue['asset_id'] = assetid
                issue['object_meta'] = ''
                issue['type'] = 'SSL' 
                findings.append(issue)
    return findings



import sys
import os
import subprocess
import tempfile
import traceback
import re
from xml.dom import minidom
import yaml
import uuid
import json
try:
    from urllib.parse import urlparse
except (ImportError,ValueError):
    from urlparse import urlparse
import logging
import shutil

class SingleQuoted(str):
  pass

def represent_single_quoted(dumper, data):
  return dumper.represent_scalar(yaml.resolver.BaseResolver.DEFAULT_SCALAR_TAG,
      data, style="'")

yaml.add_representer(SingleQuoted, represent_single_quoted)

CLEANR = re.compile('<.*?>')

def _get_rating(risk):
    if risk == '0':
        return '1'
    if risk == '1':
        return '2'
    if risk == '2':
        return '3'
    if risk == '3' or risk == '4':
        return '5'

def get_all_text_json(instance):
    itext = ""
    for key in instance:
        itext += key + ": " + instance[key] + "\n"
    return itext

def get_all_text(node):
        if node.nodeType ==  node.TEXT_NODE:
            return node.data.strip()
        else:
            text_string = ""
            for child_node in node.childNodes:
                alltext = get_all_text(child_node)
                if alltext != '':
                    text_string = text_string + alltext + '\n'
            return text_string

def run_zap(args, assetid):
    findings = []

    zapcli = shutil.which("zaproxy")
    if not zapcli:
        zapcli = shutil.which("zap.sh")
    
    if not zapcli:
        logging.error("'zaproxy' or 'zap.sh' CLI not found")
        logging.error("Please make sure zaproxy command is installed and in your $PATH")
        return findings

    report_file = None
    zap_cmd = zapcli + " -cmd -port 8081 "
    if args.planfile != None:
        logging.info("Starting web application scan for plan "+args.planfile)
        with open(args.planfile) as f:
            plan = yaml.safe_load(f)
            for e in plan['jobs']:
                if 'reportDir' in e['parameters']:
                    report_file = tempfile.NamedTemporaryFile(suffix='.json')
                    report_file = report_file.name
                    e['parameters']['reportDir'] = SingleQuoted(os.path.dirname(report_file))
                    e['parameters']['template'] = SingleQuoted("traditional-json")
                    e['parameters']['reportFile'] = SingleQuoted(os.path.basename(report_file))
            newplanfile = tempfile.NamedTemporaryFile()
            with open(newplanfile.name, "w") as npf:
                yaml.dump(plan, npf, default_flow_style=False, encoding='utf-8')
            zap_cmd = zap_cmd + " -autorun " + newplanfile.name + " 2>/dev/null"
    elif args.url != None:
        logging.info("Starting web application scan for "+args.url)
        z_path = tempfile.NamedTemporaryFile()
        zap_cmd = zap_cmd + " -quickurl "+args.url+" -quickout "+z_path.name+" 2>/dev/null"
        report_file = z_path.name

    try:
        logging.debug("ZAP command [%s]", zap_cmd)
        out = subprocess.check_output([zap_cmd], shell=True)
        out = out.decode(args.encoding)
    except subprocess.CalledProcessError:
        logging.error("Error running zaproxy...unable to run web app scan")
        logging.error(traceback.format_exc())
        return findings

    if report_file == None:
        logging.error("Report file not available. Nothing to process")
        return findings

    fp = open(report_file, mode='r')
    if fp == None:
        return findings
    contents = fp.read()
    fp.close()

    if args.planfile != None:
        jsondoc = None
        jsondoc = json.loads(contents)
        for site in jsondoc['site']:
            for alert in site['alerts']:
                rating = alert['riskcode']
                confidence = alert['confidence']
                issue = {}
                pid = alert['pluginid']
                issue['twc_id'] = 'zap-'+pid
                issue['twc_title'] = alert['name']
                issue['rating'] = _get_rating(rating)
                issue['asset_id'] = assetid
                issue['object_id'] = args.url 
                issue['object_meta'] = args.url 
                issue['type'] = 'DAST'
                desc = alert['desc']
                solution = alert['solution']
                if solution != None:
                    solution = solution.strip()
                else:
                    solution = ''
                if solution != '':
                    desc = desc + '\nSolution:\n' + solution + '\n'
                refs = alert['reference']
                if refs != None:
                    refs = refs.strip()
                else:
                    refs = ''
                if refs != '':
                    refs = refs.replace('http','\nhttp')
                    desc = desc + '\nReferences:\n' + refs + '\n' 
                instances = alert['instances']
                if len(instances) > 0:
                    desc = desc + '\nDetailed findings:\n'
                for i in instances:
                    desc = desc + '\n' + get_all_text_json(i) + '\n'
                desc = re.sub(CLEANR, '', desc)
                issue['details'] = desc
                findings.append(issue)
    else:
        xmldoc = None
        try:
            xmldoc = minidom.parseString(contents)
        except Exception:
            logging.error("Unable to parse %s", report_file)
            logging.error(traceback.format_exc())
            return findings
    
        allitems = xmldoc.getElementsByTagName('alertitem')
        for item in allitems:
            rating = item.getElementsByTagName('riskcode')[0].firstChild.data.strip()
            confidence = item.getElementsByTagName('confidence')[0].firstChild.data.strip()
            issue = {}
            pid = item.getElementsByTagName('pluginid')[0].firstChild.data.strip()
            issue['twc_id'] = 'zap-'+pid
            issue['twc_title'] = item.getElementsByTagName('name')[0].firstChild.data.strip()

            issue['rating'] = _get_rating(rating)
            issue['asset_id'] = assetid
            issue['object_id'] = args.url 
            issue['object_meta'] = args.url 
            issue['type'] = 'DAST'
            desc = item.getElementsByTagName('desc')[0].firstChild.data.strip()
            solution = item.getElementsByTagName('solution')[0].firstChild
            if solution != None:
                solution = solution.data.strip()
            else:
                solution = ''
            if solution != '':
                desc = desc + '\nSolution:\n' + solution + '\n'
            refs = item.getElementsByTagName('reference')[0].firstChild
            if refs != None:
                refs = refs.data.strip()
            else:
                refs = ''
            if refs != '':
                refs = refs.replace('http','\nhttp')
                desc = desc + '\nReferences:\n' + refs + '\n' 
            instances = item.getElementsByTagName('instances')
            if len(instances) > 0:
                desc = desc + '\nDetailed findings:\n'
            for i in instances:
                desc = desc + '\n' + get_all_text(i) + '\n'
            desc = re.sub(CLEANR, '', desc)
            issue['details'] = desc
            findings.append(issue)

    return findings



import sys
import logging
import json
import traceback
import ssl
import json
from pyVim.connect import SmartConnectNoSSL
from pyVmomi import vim

# Method that populates objects of type vimtype
def get_all_objs(content, vimtype):
    obj = {}
    container = content.viewManager.CreateContainerView(content.rootFolder, vimtype, True)
    for managed_object_ref in container.view:
        obj.update({managed_object_ref: managed_object_ref.name}),
    return obj

def discover(host, user, password):
    si = SmartConnectNoSSL(host=host, user=user, pwd=password)
    content = si.content

    #all_esxs = get_all_objs(content, [vim.HostSystem])

    #print('All ESX info')
    #for esx in all_esxs:
    #    summary = esx.summary.config
    #    print(summary.name)
    #    print(summary.product.instanceUuid)
    #    print(summary.product.productLineId)
    #    print(summary.product.version)
    #    print(summary.product.osType)
    #    print(summary.product.vendor)
    #    print('')

    data = {}
    data['id'] = content.about.instanceUuid
    data['product_id'] = content.about.productLineId
    data['vendor'] = content.about.vendor
    data['product_name'] = content.about.name
    data['product_full_name'] = content.about.fullName
    data['version'] = content.about.version
    data['os_type'] = content.about.osType

    data['licenses'] = {}
    for l in content.licenseManager.licenses:
        data['licenses']['license_name'] = l.name

    children = content.rootFolder.childEntity
    for child in children:  # Iterate though DataCenters
        datacenter = child
        data[datacenter.name] = {}  # Add data Centers to data dict
        clusters = datacenter.hostFolder.childEntity
        for cluster in clusters:  # Iterate through the clusters in the DC
            # Add Clusters to data dict
            data[datacenter.name][cluster.name] = {}
            hosts = cluster.host  # Variable to make pep8 compliance
            for host in hosts:  # Iterate through Hosts in the Cluster
                summary = host.summary.config
                hostname = summary.name
                # Add VMs to data dict by config name
                if hostname not in data[datacenter.name][cluster.name]:
                    data[datacenter.name][cluster.name][hostname] = {}
                data[datacenter.name][cluster.name][hostname]['id'] = summary.product.instanceUuid 
                data[datacenter.name][cluster.name][hostname]['product_id'] = summary.product.productLineId
                data[datacenter.name][cluster.name][hostname]['vendor'] = summary.product.vendor
                data[datacenter.name][cluster.name][hostname]['product_name'] = summary.product.name
                data[datacenter.name][cluster.name][hostname]['product_full_name'] = summary.product.fullName
                data[datacenter.name][cluster.name][hostname]['version'] = summary.product.vendor
                data[datacenter.name][cluster.name][hostname]['os_type'] = summary.product.osType
                #vms = host.vm
                #for vm in vms:  # Iterate through each VM on the host
                #    vmname = vm.summary.config.name
                #    data[datacenter.name][cluster.name][hostname][vmname] = {}
                #    summary = vmsummary(vm.summary, vm.guest)
                #    vm2dict(datacenter.name, cluster.name, hostname, vm, summary)

    print json.dumps(data, indent=4)

if __name__ == "__main__":
    discover(sys.argv[1], sys.argv[2], sys.argv[3])



import sys
import json
import os
import requests
import re
import logging
import random
import time

from . import utils

def get_machines(args, token):
    headers = { "Content-Type":"application/json", "Accept": "application/json", "Authorization": "Bearer %s" % token }
    if args.all:
        url = "https://api.securitycenter.microsoft.com/api/machines?$filter=healthStatus+eq+'Active'"
    else:
        url = "https://api.securitycenter.microsoft.com/api/machines?$filter=healthStatus+eq+'Active'+and+riskScore+eq+'High'"

    logging.info("Retrieving inventory o365") 

    resp = requests.get(url, headers=headers)
    if resp.status_code != 200:
        logging.error("Error could not get machine inventory details from O365")
        logging.error("Response content: %s" % resp.text)
        utils.tw_exit(1)
    response = resp.json()
    allmachines = response['value']
    logging.debug("Retrieved "+str(len(allmachines))+" machine details")
    assets = []
    for machine in allmachines:
        mstr = json.dumps(machine, indent=4)
        logging.debug("Processing machine")
        logging.debug(mstr)
        if machine['onboardingStatus'] != 'Onboarded':
            continue
        if machine.get('osPlatform') is None or 'Windows' not in machine['osPlatform']:
            continue
        asset = {}
        asset['owner'] = args.handle
        asset['id'] = machine['id']
        asset['name'] = machine['computerDnsName']
        asset['type'] = 'Windows'
        asset_tags = []
        asset_tags.append('Windows')
        asset_tags.append('SOURCE:O365')
        asset_tags.append('O365')
        asset_tags.append('OS_RELEASE:' + machine['osPlatform'])
        asset_tags.append(machine['osPlatform'])
        asset_tags.append('OS_RELEASE_ID:' + machine['version'])
        # Defender does not provide complete build number like '10.0.14393.3686' instead it only provides partial value like '14393'
        #asset_tags.append('OS_VERSION:' + 'Build '+str(machine['osBuild']))
        asset_tags.append('OS_ARCH:' + machine['osArchitecture'] + ' ' + machine['osProcessor'] + '-based PC')
        for tag in machine['machineTags']:
            asset_tags.append(tag)
        if asset['name'].startswith('lap'):
            asset_tags.append('LAPTOP')
        if asset['name'].startswith('wks'):
            asset_tags.append('WORKSTATION')
        if asset['name'].startswith('wow'):
            asset_tags.append('WORKSTATION_ON_WHEELS')
        if asset['name'].startswith('sh'):
            asset_tags.append('SERVER')
        asset['tags'] = asset_tags

        products = []
        logging.debug("Getting product info for "+asset['name'])
        url = "https://api.securitycenter.microsoft.com/api/machines/"+machine['id']+"/software"
        resp = requests.get(url, headers=headers)
        if resp.status_code != 200:
            logging.error("Error could not get software inventory details from O365 for machine [%s]" % machine['id'])
            logging.error("Response content: %s" % resp.text)
            allproducts = []
        else:
            allproducts = resp.json()['value']
        for product in allproducts:
            newproduct = product['vendor'] + " " + product['name']
            newproduct = newproduct.replace('_',' ')
            products.append(newproduct)
        asset['products'] = products

        # get vulnerabilities for machine id
        impacts = []
        logging.debug("Getting vulnerabilities for "+asset['name'])
        url = "https://api.securitycenter.microsoft.com/api/vulnerabilities/machinesVulnerabilities?$filter=machineId+eq+'"+machine['id']+"'"
        resp = requests.get(url, headers=headers)
        if resp.status_code != 200:
            logging.error("Error could not get vulnerability details from O365 for machine [%s]" % machine['id'])
            logging.error("Response content: %s" % resp.text)
            allvulns = []
        else:
            allvulns = resp.json()['value']
        for v in allvulns:
            finding = {}
            finding['type'] = 'IMPACT'
            finding['id_str'] = v['cveId']
            finding['percentage'] = 100
            prod = v['productVendor'] + ' ' + v['productName'] + ' ' + v['productVersion']
            prod = prod.replace('_',' ')
            finding['keyword'] = prod
            finding['product'] = prod
            finding['analysis'] = ''
            if v['fixingKbId'] != None:
                reco = 'Requires Windows patch/KB: '+v['fixingKbId']
            else:
                reco = 'Please look for patches/remediations for "'+prod+'" in related advisories' 
            finding['recommendation'] = reco 
            impacts.append(finding)
        if len(impacts) == 0:
            logging.debug("No vulnerabilities for "+asset['name'])
        else:
            asset['impacts'] = impacts

        assets.append(asset)

        r = random.uniform(0.5,1.5)
        time.sleep(r)
    return assets

def get_access_token(args):
    aad_id = args.tenant_id
    aad_app_id = args.application_id
    app_key = args.application_key
    url = "https://login.microsoftonline.com/" + aad_id + "/oauth2/token"
    resource = "https://api.securitycenter.microsoft.com"

    logging.info("Getting access token for resource [%s]...", resource) 

    resp = requests.post(url, data={"grant_type":"client_credentials", "client_id": aad_app_id, "client_secret": app_key, "resource":resource})
    if resp.status_code == 200:
        response = resp.json()
        token = response['access_token']
    else:
        logging.error("Error unable to get access token for API calls")
        logging.error("Response content: %s" % resp.text)
        utils.tw_exit(1)

    return token

# Main entry point
def get_inventory(args):
    token = get_access_token(args)
    if token is None:
        logging.error("Error unable to get access token for API calls")
        return
    args.no_scan = True # we don't want to run Impact assessment for O365 assets
    return get_machines(args,token)




import sys
import os
import logging
import re
from . import oci_utils

def get_asset_type(os_family):
    # Oracle OS Mgmt Hub only supports Oracle Linux and Windows 
    if os_family.startswith('ORACLE_LINUX_'):
        return 'Oracle Linux'
    elif os_family.startswith('WINDOWS_'):
        return 'Windows'

def get_os_release(os_family):
    return os_family.replace('_', ' ').title()

def get_inventory(args):
    assets = []
    oci_utils.set_encoding(args.encoding)

    compartment_name_dict = oci_utils.get_compartment_name_dict(args)
    compartments = oci_utils.get_compartments(args)
    for compartment in compartments:
        logging.info("Processing compartment [%s]", compartment_name_dict[compartment])
        instances_json = oci_utils.run_oci_cmd("os-management-hub managed-instance list --compartment-id '%s' --all" % compartment, args)
        instances_json = instances_json['data']['items']
        logging.info("Found [%s] compute instances in inventory", len(instances_json))
        for instance in instances_json:
            asset_dict = {}
            asset_dict['id'] = instance['id']
            asset_dict['name'] = instance['display-name']
            asset_dict['type'] = get_asset_type(instance['os-family'])
            asset_dict['owner'] = args.handle
            os_release = get_os_release(instance['os-family'])
            asset_tags = [ os_release, "OS_RELEASE:"+os_release, "OCI", "SOURCE:OCI" ]
            if args.enable_tracking_tags:
                asset_tags.append("COMPARTMENT:" + compartment_name_dict[instance['compartment-id']])
            asset_dict['tags'] = asset_tags
            products = []
            if asset_dict['type'] == "Oracle Linux":
                packages_json = oci_utils.run_oci_cmd("os-management-hub managed-instance list-installed-packages --managed-instance-id '%s' --all" % instance['id'], args)
                packages_json = packages_json['data']['items']
                for package in packages_json:
                    pname = "%s %s.%s" % (package['name'], package['version'], package['architecture'].lower())
                    products.append(pname)
            if asset_dict['type'] == "Windows":
                # OCI OS Mgmt Hub does not provide installed products for Windows OS

                updates_json = oci_utils.run_oci_cmd("os-management-hub managed-instance list-installed-windows-updates --managed-instance-id '%s' --all" % instance['id'], args)
                updates_json = updates_json['data']['items']
                patches = []
                for update in updates_json:
                    kb_nos = re.findall(r'KB[0-9]+', update['name'])
                    if len(kb_nos) == 0:
                        continue
                    patch_dict = {
                            'url': '',
                            'id': kb_nos[0],
                            'product': '',
                            'description': update['name']
                            }
                    patches.append(patch_dict)
                asset_dict['patches'] = patches
            asset_dict['products'] = products
            assets.append(asset_dict)
    logging.info("Completed inventory collection...")
    return assets




import sys
import os
import winrm
import csv
import ipaddress
import getpass
import base64
import json
import logging
import uuid
import tempfile
import shutil
import warnings
with warnings.catch_warnings():
   warnings.simplefilter("ignore", category=Warning)
   from cryptography.hazmat.backends import default_backend
   from cryptography.hazmat.primitives import hashes
   from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
   from cryptography.fernet import Fernet

def get_inventory(args):
    host_list_file = args.host_list
    if host_list_file is not None:
        with open(host_list_file, mode='r') as csv_file:
            csv_reader = csv.DictReader(csv_file, quoting=csv.QUOTE_NONE, escapechar='\\')
            password = None
            remote_hosts = []
            for row in csv_reader:
                if not args.secure and row['userpwd'].startswith('__SECURE__:'):
                    if args.password is None:
                        if password is None:
                            password = getpass.getpass(prompt="Enter password: ")
                            password = password.encode()
                    else:
                        password = args.password.encode()
                    salt = base64.b64encode(password)
                    kdf = PBKDF2HMAC(
                            algorithm=hashes.SHA256(),
                            length=32,
                            salt=salt,
                            iterations=100000,
                            backend=default_backend())
                    key = base64.urlsafe_b64encode(kdf.derive(password))
                    f = Fernet(key)
                    try:
                        epass = row['userpwd'].replace('__SECURE__:','')
                        row['userpwd'] = f.decrypt(epass.encode('utf-8'))
                    except:
                        logging.error("Failed to decrypt login details for "+row['hostname'])
                        return None
                elif row['userpwd'] != '' and not args.secure:
                    logging.warning('Unsecure login information in file. Use --secure to encrypt.')
                if '-' in row['hostname'] or '/' in row['hostname']: # IP range or CIDR is specified, then expand it
                    if '-' in row['hostname']:
                        iprange = row['hostname']
                        iprange = iprange.replace(' ','')
                        tokens = iprange.split('-')
                        if len(tokens) != 2 or len(tokens[0])==0 or len(tokens[1])==0:
                            logging.error("Skipping invalid range [%s]", row['hostname'])
                            continue
                        logging.info("Enumerating IPs based on specified range [%s]", iprange)
                        try:
                            if sys.version_info[0] < 3:
                                startip = ipaddress.IPv4Address(unicode(tokens[0]))
                                endip = ipaddress.IPv4Address(unicode(tokens[1]))
                            else:
                                startip = ipaddress.IPv4Address(tokens[0])
                                endip = ipaddress.IPv4Address(tokens[1])
                            cidrs = []
                            cidrs = [ipaddr for ipaddr in ipaddress.summarize_address_range(startip,endip)]
                        except Exception as e:
                            logging.error("Encountered exception: %s",e)
                            logging.error("Error converting IP range [%s] to CIDRs. Skipping it...", iprange)
                            continue
                        logging.info("Converted IP range [%s] to CIDRs %s", iprange, cidrs)
                    if '/' in row['hostname']:
                        logging.info("Enumerating IPs based on specified CIDR [%s]", row['hostname'])
                        try:
                            if sys.version_info[0] < 3:
                                cidrs = [ipaddress.ip_network(unicode(row['hostname'],"ascii"))]
                            else:
                                cidrs = [ipaddress.ip_network(row['hostname'])]
                        except Exception as e:
                            logging.error("Encountered exception: %s",e)
                            logging.error("Invalid CIDR [%s] specified. Skipping it...", row['hostname'])
                            continue
                    for cidr in cidrs:
                        for a in cidr:
                            trow = row.copy()
                            trow['hostname'] = str(a)

                            # Remove hard-coded asset ID and name for CIDR, as it will overwrite same asset
                            # These will based on host IP address automatically
                            trow['assetname'] = None

                            remote_hosts.append(trow)
                            remote_hosts[-1]['remote'] = True
                            logging.info("Enumerated IP: %s", a)
                else:
                    remote_hosts.append(row)
                    remote_hosts[-1]['remote'] = True

        if args.secure:
            # secure the host list
            logging.info("Securing host list file")
            if args.password is None:
                pp1 = getpass.getpass(prompt="Enter password: ")
                pp2 = getpass.getpass(prompt="Re-enter password: ")
                if pp1 != pp2:
                    logging.info("Passwords don't match. Try again.")
                    return None
            else:
                pp1 = args.password
            password = pp1.encode()
            salt = base64.b64encode(password)
            kdf = PBKDF2HMAC(
                    algorithm=hashes.SHA256(),
                    length=32,
                    salt=salt,
                    iterations=100000,
                    backend=default_backend())
            key = base64.urlsafe_b64encode(kdf.derive(password))
            f = Fernet(key)
            # verify the key if possible
            with open(host_list_file, mode='r') as csv_file:
                csv_reader = csv.DictReader(csv_file, quoting=csv.QUOTE_NONE, escapechar='\\')
                for row in csv_reader:
                    try:
                        if row['userpwd'] != '' and row['userpwd'].startswith('__SECURE__:'):
                            epass = row['userpwd'].replace('__SECURE__:','')
                            epass  = f.decrypt(epass)
                    except:
                        logging.error("Invalid password")
                        logging.error("Please use the same password as was used previously to secure the file")
                        return None
            # create new file with secured information. This will be renamed subsequently
            new_csv_file = uuid.uuid4().hex
            new_csv_file = tempfile.gettempdir() + os.path.sep + new_csv_file + ".csv"
            if os.path.isfile(new_csv_file):
                os.remove(new_csv_file)
            try:
                # secure the new rows in the file
                with open(new_csv_file, mode='w') as csvfile:
                    fieldnames = ['hostname','userlogin','userpwd']
                    writer = csv.DictWriter(csvfile, fieldnames=fieldnames, quoting=csv.QUOTE_NONE, escapechar='\\')
                    writer.writeheader()
                    for h in remote_hosts:
                        if h['userpwd'] != '' and not h['userpwd'].startswith('__SECURE__:'):
                            h['userpwd'] = '__SECURE__:'+f.encrypt(h['userpwd'].encode('utf-8')).decode('utf-8')
                        del h['remote']
                        writer.writerow(h)
            except Exception as err:
                logging.error("Unable to save secured CSV file")
                logging.error("%s", err)
                os.remove(new_csv_file)
                utils.tw_exit(1)
            os.remove(host_list_file)
            shutil.copyfile(new_csv_file, host_list_file)
            os.remove(new_csv_file)
            logging.info("Host list file secured")
            return None
        else:
            return discover_windows_hosts(args, remote_hosts)

def discover_windows_hosts(args, hosts):
    assets = []
    for host in hosts:
        asset = discover_windows_host(args, host)
        if asset is not None:
            assets.append(asset)
    return assets

def discover_windows_host(args, host):
    logging.info("Running discovery for host [%s]", host['hostname'])
    winrm_session, asset_id = open_winrm_session(host, args)
    if asset_id is None:
        return None
    asset_id = asset_id.replace('/','-')
    asset_id = asset_id.replace('/','-')
    asset_name = asset_id
    logging.info("Getting OS details")
    temp_out = run_ps_command(args, winrm_session,
     """$computer_info = Get-ComputerInfo
        Write-Host $computer_info.OsName
        Write-Host $computer_info.OsVersion
        Write-Host $computer_info.CsSystemType
        Write-Host $computer_info.OsArchitecture""")
    if temp_out is None:
        return None
    temp_out = temp_out.split('\n')
    if len(temp_out) != 4:
        return None
    os_name = temp_out[0]
    os_version = temp_out[1]
    mc_arch = temp_out[2]
    bit_arch = temp_out[3]
    os_arch = bit_arch + ' ' + mc_arch
    os_version_ubr = run_ps_command(args, winrm_session,"""$os_version_ubr = (Get-ItemProperty 'HKLM:\SOFTWARE\Microsoft\Windows NT\CurrentVersion').UBR
        Write-Host $os_version_ubr""")
    if os_version_ubr is not None and os_version_ubr != "":
        os_version_tokens = os_version.split(' ')
        os_version = os_version_tokens[0] + '.' + os_version_ubr
        for index in range(1,len(os_version_tokens)):
            os_version = os_version + ' ' + os_version_tokens[index]
    os_release_id = run_ps_command(args, winrm_session,"""$os_release_id = (Get-ItemProperty 'HKLM:\SOFTWARE\Microsoft\Windows NT\CurrentVersion').ReleaseId
        Write-Host $os_release_id""")

    logging.info("Getting patch information")
    temp_out = run_ps_command(args, winrm_session,"""$hot_fixes = Get-HotFix
        $patch_json_array = New-Object System.Collections.Generic.List[System.Object]
        foreach ($hot_fix in $hot_fixes) { $patch_entry_json = @{id=$hot_fix.HotFixID}
        $patch_json_array.add($patch_entry_json)}
        $patch_json = $patch_json_array | ConvertTo-Json;$patch_json_str = $patch_json.ToString()
        Write-Host $patch_json_str""")
    try:
        patches = json.loads(temp_out)
    except Exception as e:
        logging.error("Error parsing patch JSON")
        return None

    logging.info("Getting product inventory")
    temp_out = run_ps_command(args, winrm_session,"""$unique_products = New-Object System.Collections.Generic.List[string]
        $product_json_array = New-Object System.Collections.Generic.List[string]
        $temp_array = Get-ItemProperty HKLM:\\Software\Wow6432Node\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\* | Select-Object Publisher, DisplayName, DisplayVersion
        $temp_array | foreach { $var = $_ ; $product = $var.'DisplayName'; $vendor = $var.'Publisher'; $version = $var.'DisplayVersion'; if ($product -and $version) { $product_details = $product.Trim() + ' ' + $version.Trim(); $product_json_array.add($product_details); }}
        $temp_products_count = $temp_array.Length
        $temp_array = Get-ItemProperty HKLM:\\Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\* | Select-Object Publisher, DisplayName, DisplayVersion
        $temp_array | foreach { $var = $_ ; $product = $var.'DisplayName'; $vendor = $var.'Publisher'; $version = $var.'DisplayVersion'; if ($product -and $version) { $product_details = $product.Trim() + ' ' + $version.Trim(); $product_json_array.add($product_details); }}
        $ie_version = Get-ItemProperty 'HKLM:\Software\Microsoft\Internet Explorer' | Select-Object svcVersion, Version
        if ($ie_version.svcVersion -ne $null) {
            $product_json_array.add('Internet Explorer ' + $ie_version.svcVersion)
            $temp_products_count = $temp_products_count + 1
        }
        else {
            if ($ie_version.Version -ne $null) {
                $product_json_array.add('Internet Explorer ' + $ie_version.Version)
                $temp_products_count = $temp_products_count + 1
            }
        }
        $temp_products_count = $temp_products_count + $temp_array.Length
        $temp_array = Get-WMIObject -ClassName Win32_Product
        foreach ($row in $temp_array) { $product_details = $row.Name.Trim() + ' ' + $row.Version.Trim(); if  ($product_json_array -notcontains $product_details) { $product_json_array.Add($product_details)} }
        $product_json = $product_json_array | ConvertTo-Json
        $product_json_str = $product_json.ToString()
        Write-Host $product_json_str """)
    try:
        products = json.loads(temp_out)
    except Exception as e:
        logging.error("Error parsing product JSON")
        return None

    logging.info("Completed discovery of host")
    asset_data = {}
    asset_data['id'] = asset_id
    asset_data['name'] = asset_name
    asset_data['type'] = 'Windows'
    asset_data['owner'] = args.handle
    asset_data['products'] = products
    asset_data['patches'] = patches
    asset_tags = []
    asset_tags.append('OS_RELEASE:' + os_name)
    asset_tags.append(os_name)
    asset_tags.append('OS_VERSION:' + os_version)
    if os_release_id is not None and os_release_id.strip() != "":
        asset_tags.append('OS_RELEASE_ID:' + os_release_id)
    if os_arch.strip() != "":
        asset_tags.append('OS_ARCH:' + os_arch)
    asset_tags.append('Windows')
    asset_data['tags'] = asset_tags
    
    return asset_data

def open_winrm_session(host, args):
    winrm_session = None
    asset_id = None
    auth_modes = ['kerberos', 'ntlm', 'credssp', 'plaintext']
    for auth_mode in auth_modes:
        try:
            logging.debug("Attempting to authenticate using [%s]", auth_mode)
            winrm_session = winrm.Session(host['hostname'], auth=(host['userlogin'], host['userpwd']), transport=auth_mode)
        except Exception as e:
            logging.debug("Got exception. Unable to authenticate to host using [%s]", auth_mode)
            logging.debug(e)
            continue
        asset_id = run_ps_command(args, winrm_session, "Write-Host $env:ComputerName", True)
        if asset_id is None:
            logging.debug("Unable to authenticate to host using [%s]", auth_mode)
            continue
        else:
            logging.debug("Authenticated to host using [%s]", auth_mode)
            break
    if asset_id is None:
        logging.info("Unable to authenticate to host [%s]", host['hostname'])
        return None, None
    else:
        return winrm_session, asset_id

def run_ps_command(args, winrm_session, command, no_err_msg=False):
    try:
        cmd_out = winrm_session.run_ps(command)
        cmd_out = cmd_out.std_out.decode(args.encoding).strip()
        return cmd_out
    except Exception as e:
        if not no_err_msg:
            logging.error("Error running command on host: %s", str(e))
        logging.debug("Commmand: %s", command)
        return None
